﻿if (Test-Path .\images.drv) {
    $global:scripts_drive=Get-Content .\images.drv
} else {
    $global:scripts_drive="Z:"
}

#Function - Returns firmware mode
function get_firmware {
	$temp=bcdedit
	if ($temp -match ".efi") {
		return "UEFI"
	} elseif ($temp -match "winload") {
		return "BIOS"
	} else {
		return "None"
	}
}

#Function - Sets global vars for PC hardware. All spaces in vars are removed.
function get_hw_info {
	$global:serial=((wmic bios get serialnumber) -split '\n')[2] -replace '\s',''
	$global:vendor=((wmic csproduct get vendor) -split '\n')[2] -replace '\s',''
	$global:model=((wmic csproduct get name) -split '\n')[2] -replace '\s',''
	$global:version=((wmic csproduct get version) -split '\n')[2] -replace '\s',''
	$global:firmware=get_firmware 

    #Account for Lenpovo putting model name in version field    
    if ($vendor -match "LENOVO") {
        $model=$version
    }

}

#Function - Returns current date/time stamp
function date_stamp {
	return (get-date -format "MM/dd/yyyy HH:mm")
}

#Function - Read setting from ini file
#https://devblogs.microsoft.com/scripting/use-powershell-to-work-with-any-ini-file/
#Usage => $iniContent = Get-IniContent "path\to\ini\file"
#Usage => iniContent["section"].value
function Get-IniContent ($FilePath) 
{
    $ini = @{}
    switch -regex -file $FilePath
    {
        "^\[(.+)\]" # Section
        {
            $section = $matches[1]
            $ini[$section] = @{}
            $CommentCount = 0
        }
        "^(;.*)$" # Comment
        {
            $value = $matches[1]
            $CommentCount = $CommentCount + 1
            $name = "Comment" + $CommentCount
            $ini[$section][$name] = $value
        }
        "(.+?)\s*=(.*)" # Key
        {
            $name,$value = $matches[1..2]
            $ini[$section][$name] = $value
        }
    }
    return $ini
}

#Function - Read config settings from config.ini into global vars
function read_config {
	$iniContent = Get-IniContent "config.ini"
	$global:server = $iniContent["share"].server
	$global:share = $iniContent["share"].share
	$global:domain = $iniContent["share"].domain
	$global:username = $iniContent["share"].username
	$global:password = $iniContent["share"].password
	$global:image_path = $iniContent["path"].image_path
	$global:driver_path = $iniContent["path"].driver_path
	$global:msi_path = $iniContent["path"].msi_path
	$global:script_path = $iniContent["path"].script_path
	$global:config_path = $iniContent["path"].config_path
	$global:log_path = $iniContent["path"].log_path
	$global:history_path = $iniContent["path"].history_path
	$global:util_path = $iniContent["path"].util_path
	$global:extra_path = $iniContent["path"].extra_path
}

#Function - Get drive letter of Windows volume
function get_Windows_drive {
    $letter = ((get-psdrive |where {$_.Root -match ":"} |% {if (Test-Path ($_.Root + "Windows")){$_.Root}}).Substring(0,1)).Split('\n') | select -First 1
    if (($letter -match "X") -and ($firmware -match "UEFI")) {
        Add-Content x:\imaging\win_letter.txt "select disk 0"
        Add-Content x:\imaging\win_letter.txt "select partition 3"
        Add-Content x:\imaging\win_letter.txt "assign letter=W"
        diskpart /s x:\imaging\win_letter.txt | Out-Null
        $letter="W"
    }
    return $letter
}

#Function - Generate and return image name based on OS on c: drive, e,g. Windows10ProN_x64_20H2_01.01.2021_17.59
#HKLM\SOFTWARE\Microsoft\Windows NT\CurrentVersion\ProductName - Windows 10 Pro N
#HKLM\SOFTWARE\Microsoft\Windows NT\CurrentVersion\DisplayVersion - 20H2
#HKLM\SOFTWARE\Microsoft\Windows NT\CurrentVersion\BuildLabEx - 19041.1.amd64fre.vb_release.191206-1406
function get_image_name {
    $drive=get_Windows_drive
    reg load HKLM\OldRegistry ($drive + ":\windows\system32\config\software") | Out-Null
    $OS_release=((Get-ItemProperty -Path "HKLM:\OldRegistry\Microsoft\Windows NT\CurrentVersion").BuildLabEx).split(".")[3]
    $OS_ver=(Get-ItemProperty -Path "HKLM:\OldRegistry\Microsoft\Windows NT\CurrentVersion").ProductName -replace "\s",""

    #Windows 11 still reports Windows 10 for product name so have to account for this
    if ($OS_release -match "co_release") {
        $OS_ver=$OS_ver -replace '10','11'
    }

	$OS_build=(Get-ItemProperty -Path "HKLM:\OldRegistry\Microsoft\Windows NT\CurrentVersion").DisplayVersion
	$OS_arch_temp=((Get-ItemProperty -Path "HKLM:\OldRegistry\Microsoft\Windows NT\CurrentVersion").BuildLabEx).split(".")[2]
	reg unload HKLM\OldRegistry | Out-Null

	if ($OS_arch_temp -match "64") {
		$OS_arch = "x64"
	} else {
		$OS_arch = "x86"
	}

	$current_date = (get-date -format "MM.dd.yyyy_HH.mm")
	$filename=($OS_ver + "_" + $OS_arch + "_" + $OS_build + "_" + $current_date)

	return $filename
}

#Function - Capture image
function capture_image {
    $image_name=get_image_name
	Write-Output ""
	Write-Output "Requesting info for image capture"
    Write-Output  ("Filename for image will be $image_name" + "_comment.wim")
	Write-Output "Spaces in comment will be converted to _ character"

    Do {
        $comment = Read-Host -Prompt "Enter comment to include in image file name"        
    } until (-Not ([string]::IsNullOrWhiteSpace($comment)))

    $image_name=($image_name + "_" + $comment + ".wim") -replace "\s","_"
    $image_desc= Read-Host "Enter a description for the image to be captured (alphanumeric characters only)"
    
    if ([string]::IsNullOrWhiteSpace($image_desc)) {
        Write-Output "No description specified. Using comment for description."
        $image_desc = $comment
    }
    
    $windows_drive=((get_Windows_drive) + ":\")
    Dism /Capture-Image /ImageFile:"$image_path\$image_name" /CaptureDir:$windows_drive /Name:"$image_desc"
	Add-Content "$image_path\image_list.txt" ($image_name + "," + $image_desc)
	
	Write-Output ""
	Write-Output "Image saved to $image_path\$image_name"
	$answer=Read-Host -Prompt "Press Return key to shut down"

    if ($answer -match "z") {
        Write-Output "Dropping to commandline for troubleshooting"
        exit
    } else {
    	wpeutil shutdown
    }
}

#Function - Get computername and assigned image based on computer serial number
#image_mappings.txt file should be in format serial,computername,image,cf1:value1,cf2:value2,cf3:value3
function get_mappings {
	#Get last matching device entry based on serial from image_mappings.txt. Last entry wins.
    $temp=get-content "$config_path\image_mappings.txt" -ReadCount 1000 | foreach { $_ -match "$serial" } | select-object -Last 1
    if ([string]::IsNullorEmpty($temp)) {
		#do nothing
	} else {
		$global:computername=$temp.split(",")[1] -replace '\s',''
		$global:image=$temp.split(",")[2] -replace '\s',''
	}
}

#Function - Edit registry to implement settings
function edit_registry {
    #No need to load registry since function called before this, install_client, has already loaded registry
    #reg load HKLM\NewRegistry w:\windows\system32\config\SOFTWARE | Out-Null

    New-Item -Path HKLM:\NewRegistry\PSImage

    $client_entry=get-content "$config_path\image_mappings.txt" -ReadCount 1000 | foreach { $_ -match "$serial" } | select-object -Last 1

    if (-Not ( $client_entry-eq $null)) {
        $cf_count=$client_entry.split(",").count

        if ( [int]$client_id -gt 0 ) {
            #Since client.id available we know client was imaged via fileset so device entry already exists 
            #Write custom fields immediately to server instead of passing to registry and waiting to do it in post-imaging via fileset
            if ($cf_count -gt 3){
                Write-Output "Found client_id => imaging triggered via fileset. Custom fields present for device in image_mappings.txt. Setting custom fields on server now."
                Write-Output ""

                for ($i=3; $i -lt $cf_count; $i+=1) {
                    $cf_field=$client_entry.split(",")[$i]
                    
                    if ( $cf_field -match ":" ) {
                        $cf_name=$cf_field.split(":")[0]
                        $cf_value=$cf_field.split(":")[1]

                        write_custom_field $cf_name "$cf_value"
                    }
                }
            } else {
                Write-Output "Found client_id => imaging triggered via fileset. No custom fields for device in image_mappings.txt."
                Write-Output ""
            }
        } else {
            #No client.id present so have to write custom field values to registry to be applied to server later via fileset         
            #Write custom fields to registry only if any are present
            if ($cf_count -gt 3) {
                for ($i=3; $i -lt $cf_count; $i+=1) {
                    $cf_temp=$client_entry.split(",")[$i]
                    if ( $cf_temp -match ":" ) {
                        $custom_fields=$custom_fields + "," + $cf_temp
                    }
                }
                $custom_fields=$custom_fields.substring(1)

                Write-Output "Found custom fields. Writing custom field data ""$custom_fields"" to HKLM\Software\PSImage\CustomFields registry value."     
                Write-Output ""
                New-ItemProperty -Path HKLM:\NewRegistry\PSImage -Name "CustomFields" -Value "$custom_fields"
            } else {
                Write-Output "No custom fields found for this device."
                Write-Output ""
            }
        }
    }

    #Write start time and image name to registry
    Write-Output "Writing imaging date/time and image name to PC registry at HKLM\Software\PSImage"
    Write-Output ""
    $image_date=date_stamp
    New-ItemProperty -Path HKLM:\NewRegistry\PSImage -Name "ImageBegin" -Value "$image_date"
    New-ItemProperty -Path HKLM:\NewRegistry\PSImage -Name "ImageName" -Value "$image"

    Write-Output "Unloading registry. May get Access Denied. This is OK."
    Write-Output ""
	Start-Sleep -s 5
    #reg unload often fails with access denied for some reason so don't bother
    reg unload HKLM\NewRegistry | Out-Null
}

#Function - Copy FW client to c:\Windows\Temp\psimage and prep PC to a) be renamed and b) install the FW client
function install_client {
    Write-Output ""
    Write-Output "Creating psimage folder. Prepping for computername change and FW client installation."
    mkdir w:\Windows\Temp\psimage

    copy $script_path\fw_exception.bat w:\Windows\Temp\psimage\
    copy $script_path\runonce.ps1 w:\Windows\Temp\psimage\
    copy $msi_path\FileWaveClient.msi w:\Windows\Temp\psimage\


    if (Test-Path "$msi_path\custom") {
        Write-Output ""
        Write-Output "Found Kiosk customization folder. Copying to reimaged OS."
        Copy-Item -Path "$msi_path\custom" -Destination "w:\Windows\Temp\psimage" -Recurse
    }
   
    $computername | Out-File w:\Windows\Temp\psimage\computername.txt
    $iniContent = Get-IniContent "config.ini"
    $autousername=$iniContent["autologin"].autousername
    $autopassword=$global:autopassword = $iniContent["autologin"].autopassword

	reg load HKLM\NewRegistry w:\windows\system32\config\SOFTWARE | Out-Null
    New-ItemProperty -Path "HKLM:\NewRegistry\Microsoft\Windows NT\CurrentVersion\Winlogon" -Name "DefaultUserName" -Value "$autousername"
    New-ItemProperty -Path "HKLM:\NewRegistry\Microsoft\Windows NT\CurrentVersion\Winlogon" -Name "DefaultPassword" -Value "$autopassword"
    New-ItemProperty -Path "HKLM:\NewRegistry\Microsoft\Windows NT\CurrentVersion\Winlogon" -Name "AutoAdminLogon" -Value "1"

    New-ItemProperty -Path "HKLM:\NewRegistry\Microsoft\Windows\CurrentVersion\RunOnce" -Name "!PSImage0" -Value "c:\WINDOWS\system32\WindowsPowerShell\v1.0\powershell.exe -ExecutionPolicy Bypass -File c:\windows\temp\psimage\runonce.ps1"
    #No need to unload since next function invoked, edit_registry, will unload it
    #reg unload HKLM\NewRegistry | Out-Null
}

#Function - Install drivers for computer model
function install_drivers {
    if (Test-Path "$scripts_drive\models\$model") {
        Write-Output ""
        Write-Output "Found drivers for $model in models folder. Installing drivers."
        dism /image:W: /add-driver /driver:"$scripts_drive\models\$model" /recurse
    } else {
        Write-Output "Could not find drivers for this PC. $scripts_drive\models\$model does not exist. Skipping driver installation."
    }
}

#Function - Copy optional items to c:\windows\temp\psimage new OS (actually w:\windows\temp\psimage in WinPE) if they exist
function copy_optional_items {
    Write-Output ""
    if ((Get-ChildItem $exta_path | Measure-Object).Count -ne 0) {
        Write-Output "Items in extra folder copied to psimage folder on restored OS"
        copy $extra_path\* w:\windows\temp\psimage\
    } else {
        Write-Output "extra folder is empty so nothing to copy"
    }

    Write-Output ""
    if (Test-Path "$script_path\to_do.bat") {
        Write-Output "to_do.bat copied to psimage folder on restored OS"
        copy "$script_path\to_do.bat" w:\windows\temp\psimage\
    } else {
        Write-Output "No to_do.bat present in $script_path folder"
    }
}

#Function - Flip "To Be Reimaged" custom field to 0
function flip_custom_field {
    $windows_drive=((get_Windows_drive) + ":")

    if (Test-Path "$windows_drive\windows\temp\device.id") {

#Ignore SSL cert. Otherwise get SSL/TLS error with self-signed certs. Do not indent this block.
add-type @"
    using System.Net;
    using System.Security.Cryptography.X509Certificates;
    public class TrustAllCertsPolicy : ICertificatePolicy {
        public bool CheckValidationResult(
            ServicePoint srvPoint, X509Certificate certificate,
            WebRequest request, int certificateProblem) {
            return true;
        }
    }
"@
[System.Net.ServicePointManager]::CertificatePolicy = New-Object TrustAllCertsPolicy
#End Ignore SSL cert block.

        $global:server_dns = Get-Content -Path ($scripts_drive + "\config\server.dns")
        $global:app_token = Get-Content -Path ($scripts_drive + "\config\app.token")
        $global:api="https://" + $server_dns + "/filewave/api/devices/v1/devices/" + $client_id
        $global:header=@{Authorization=$app_token}
        $data='{"custom_fields": {"to_be_reimaged": false}}'

        Write-Output "Flipping to_be_reimaged field to false"
        Invoke-RestMethod -Method PATCH -Headers $header -Body $data -ContentType application/json -Uri $api

        Start-Sleep 3
        Write-Output ""
    }
}

#Function - Write custom fields
function write_custom_field ($cf, $val) {
    #$server_dns, $api, $header already set globally in flip_custom_field function
    $data="{""custom_fields"": {""$cf"": ""$val""}}"
    Write-Output "Setting custom field $cf to $val"
    Invoke-RestMethod -Method PATCH -Headers $header -Body $data -ContentType application/json -Uri $api
}

#Function - Restore image
function restore_image {
    $client_entry=get-content "$config_path\image_mappings.txt" -ReadCount 1000 | foreach { $_ -match "$serial" } | select-object -Last 1

    if (-Not ( $client_entry-eq $null)) { 
        $capture_mapping=$client_entry.split(",")[2]
        if ( $capture_mapping -match "capture") {
            Write-Output "Found entry in image_mappings.txt with image name set to capture. Starting image capture."
            capture_image
        }
    }

    $windows_drive=((get_Windows_drive) + ":")

    if ( Test-Path -Path "$windows_drive\Windows\Temp\client.id" ) {
        $global:client_id=Get-Content -Path ($windows_drive + "\windows\temp\client.id")
    } else {
        $global:client_id="0"
    }

    #Log reimaging operation to file \logs\restore\serialnumber_date_time.log on network share
    $current_date = (get-date -format "MM.dd.yyyy_HH.mm")
    $log_file=$serial + "_" + $current_date + ".log"
    if ( Test-Path "$history_path\restore" ) {
        #do nothing
    } else {
        New-Item -Path "$history_path" -Name "restore" -ItemType "directory"
    }
    Start-Transcript -Path "$history_path\restore\$log_file"

	get_mappings

    if ( Test-Path $image_path\$image ) {
        #image exists so do nothing
    } else {
        Write-Output "Assigned image $image cannot be found in images folder. Check image_mappings.txt for typo in entry for this device with serial number $serial."
        exit
    }

    #Flip custom field "To Be Reimaged" to false as early as possible to so device is removed from "To Be Reimaged" smart group 
    flip_custom_field
    
	if ([string]::IsNullorEmpty($computername)) {
    	Import-CSV $image_path\image_list.txt | Format-Table #List available images
		Write-Output "No imaging association currently exists for this PC"

        #Validate image name exists
        do {
            $image=Read-Host -Prompt "Enter image name to restore"
            $image=$image -replace '\s',''
        } until ((Test-Path "$image_path\$image") -and (-Not ([string]::IsNullOrWhiteSpace($image))))

		
        #Validate entered computername to make sure a) no spaces and b) < 16 characters
        do {
            $computername=Read-Host -Prompt "What do you want to name this PC (15 char alphanumeric limit, no spaces, dashes allowed)" 
            if ($computername -notmatch " " -and $computername.Length -lt 16 -and $computername.Length -gt 0) {break}
            Write-Output "Invalid computername. Name contains a space or is more than 15 characters"
        } until ($computername -notmatch " " -and $computername.Length -lt 16 -and (-Not ([string]::IsNullOrWhiteSpace($computername))))
        
        $computername=$computername -replace '\s',''

        Write-Output ""
        Write-Output "Computer model is $vendor $model with serial $serial"
        Write-Output "Specified computer name is $computername"
        Write-output "Selected image is $image"
	} else {
        Write-Output "Existing imaging assocation found for this PC"
        Write-Output "$vendor $model with serial $serial"
        Write-Output "Assigned computer name is $computername"
        Write-Output "Assigned image is $image"

        #Prompt user for image name if the one specified in image_mappings.txt is invalid
        if (Test-Path "$image_path\$image") {
            Write-Host "Found image $image in images folder"
        } else {
            Write-Host "Could not find assigned image $image in images folder"
            Write-Host "Available images listed below"
            Get-ChildItem -Path \\dellserver\psimage\images\* -Include *.wim -Name

            do {
                $image=Read-Host -Prompt "Enter image name to restore"
                $image=$image -replace '\s',''
            } until ((Test-Path "$image_path\$image") -and (-Not ([string]::IsNullOrWhiteSpace($image))))
        }
    }

	if ( $firmware -match "None") {
		Write-Output "Hard disk is either blank or corrupt so can't determine firmware mode"
		do {
			$firmware=Read-Host -Prompt "What is the firmware mode for this PC (UEFI or BIOS)"
            $firmware=$firmware.ToUpper()
            if (($firmware -match "UEFI") -Or ($firmware -match "BIOS")) {break}
            Write-Output "Invalid firmware mode. Enter either UEFI or BIOS."
		} until (($firmware -match "UEFI") -Or ($firmware -match "BIOS"))
	}

    $start_time=date_stamp

    Write-Output ""
	if ( $firmware -match "UEFI") {
		Write-Output "Repartitioning disk for UEFI"
		diskpart /s CreatePartitions-UEFI.txt
	} elseif ( $firmware -match "BIOS") {
		Write-Output "Repartitioning disk for BIOS"
		diskpart /s CreatePartitions-BIOS.txt
	} 

    Write-Output ""
	Write-Output "Restoring disk image"
	dism /Apply-Image /ImageFile:$image_path\$image /Index:1 /ApplyDir:W:\
    Write-Output ""
	Write-Output "Copying boot files to system partition"
	W:\Windows\System32\bcdboot.exe W:\Windows /s S:
    Write-Output ""
	Write-Output "Copying Windows RE image to Windows RE Tools partition"
	md R:\Recovery\WindowsRE
	xcopy /h W:\Windows\System32\Recovery\Winre.wim R:\Recovery\WindowsRE\
    Write-Output ""
	Write-Output "Registering location of recovery tools"
	W:\Windows\System32\Reagentc.exe /Setreimage /Path R:\Recovery\WindowsRE /Target W:\Windows
	
	Start-Sleep -s 5 #Hiding recovery tools partition sometimes fails without this delay due to drive being busy
	Write-Output ""
    Write-Output "Hiding recovery tools partition"
	if ( $firmware -match "UEFI") {
		diskpart /s HideRecoveryPartitions-UEFI.txt
	} else {
		diskpart /s HideRecoveryPartitions-BIOS.txt
	}
	
    install_drivers
    install_client
	edit_registry

    copy_optional_items

    $end_time=date_stamp
    Add-Content "$history_path\history.log" "$serial,$computername,$firmware,$image,$start_time,$end_time"

    Write-Output ""
    Write-Output "Done restoring image. Imaging phase complete."
    Write-Output "Moving to post-imaging phase. Check c:\Windows\Temp\psimage.log on client device for details on post-imaging actions."

    Stop-Transcript

    #If debug is on then exit instead of reboot to troubleshoot
    if (Test-Path -Path "$config_path\debug.on") {
        Write-Output ""
        Write-Output "Debug mode is on. Dropping to command prompt to review history."
        $status="valid"
    } else {
    	wpeutil reboot
    }
}

function main_menu ($status) {
    Write-Output "---==[ PSImage Menu ]==---"
    Write-Output ""
    Write-Output "1. Capture reference image"
    Write-Output "2. Restore image"
    Write-Output "--======================--"
}

read_config
get_hw_info

clear

if (Test-Path "$config_path\mode.reimage") {
    restore_image
} else {
    do {
        $global:status="invalid"
        main_menu
        $choice1 = Read-Host -Prompt "What do you want to do"
        if ($choice1 -match "1") {
            $status="valid"
            Write-Output ""
            Write-Output "Capturing image"
            capture_image
	    } elseif ($choice1 -match "2") {
            $status="valid"
            Write-Output ""
            Write-Output "Restoring image"
		    restore_image
        } elseif ($choice1 -match "z") {
            $status="valid"
            Write-Output ""
            Write-Output "Dropping to commandline for troubleshooting"
            exit           
        } else {
		    $status="invalid"
		    Write-Output ""
            Write-Output "Invalid selection. Try again."
            Write-Output ""
        }
    } while ($status -match "invalid")
}
